from .calculus import sigmanotation,productnatation,factorial
from .combinatorics import permutation,combination,posibilty
from .errors import NonCompliancaRecognition,DefinitionError,SizeLimitExceededError,WrongDataTypeError,NonSymbolicValue,RequiredModule
from .matrix import Matrix
from .plane_geometry import Vector,Line 
from .sets import Bool_Integer,Bool_Natural,Bool_RealNumber,Integer
from .symbolic import SymbolicVariable,SymbolicMatrix,SymbolicExpression,Sequence

__version__ = "1.4.3"

__all__ = [
    "SymbolicVariable",
    "SymbolicMatrix",
    "SymbolicExpression",
    "Sequence",
    "Bool_Integer",
    "Bool_Natural",
    "Bool_RealNumber",
    "Integer",
    "Line",
    "Vector",
    "Matrix",
    "NonSymbolicValue",
    "RequiredModule",
    "WrongDataTypeError",
    "SizeLimitExceededError",
    "DefinitionError",
    "NonCompliancaRecognition",
    "posibilty",
    "sigmanotation",
    "productnatation",
    "permutation",
    "factorial",
    "combination"]





def NonCompliancaRecognition():
    """
    This error is caused by the given parameter 
    being given that does not match the definition of the given operation.

    Reasons:
    may be caused by giving an inappropriate value to the domain
    A value that is not appropriate for the given operation may have been given
    
    """

    print("[ERROR] Non-Complianca Recognition Error ")
    exit()

class DefinitionError(Exception): 
    """
    Mathematics is based on definitions.
    If you make a definitional error in the operations, 
    the whole operation will be wrong.
    This error occurs when you make the mathematical definition incorrect.
    """
    pass

class SizeLimitExceededError(Exception):
    """The operations described are only valid in 2 and 3 dimensional planes and spaces."""
    pass

class WrongDataTypeError(Exception):
    """
    An error is thrown because the data type in the given 
    parameter does not match the requested data type.
    """

    """
    def __init__(self, variable_name, expected_type):
        message = f" The variable {variable_name} must be of type {expected_type.__name__}"
        super().__init__(message)
    """
    pass

class RequiredModule(Exception):
    """
    There are external modules required for some functions.
    If the required module is not installed, you will get this error.
    """
    pass

class NonSymbolicValue(Exception):
    pass


from .errors import * 


class Matrix:
    """
    In mathematics,
    a matrix is a rectangular array of numbers or other mathematical objects
    with elements or entries arranged in rows and columns,
    usually satisfying certain properties of addition and multiplication.
    """
    def __init__(self, arrays):

        if not arrays: 
            raise DefinitionError("Data Cannot Be Empty") 
        if not isinstance(arrays, list) or not all(isinstance(row, list) for row in arrays):
            raise DefinitionError("İncorrectly Defined")
        
        self.arrays = arrays
        self.rows = len(arrays)

        try:
            self.cols = len(arrays[0])
            if not all(len(row) == self.cols for row in arrays):
                raise DefinitionError("Incorrectly Defined")
        except IndexError:
            raise DefinitionError("indentation cannot be empty. Did you mean ?: zero matrix")
        
    def get_column(self, index):
        """Returns the specified column as a list."""
        if index < 0 or index >= self.cols:
            raise IndexError("Column index out of range.")
        return [row[index] for row in self.arrays]
    
    def DelColumn(self, index): 
        """
        Allows to delete the column indexed in the given parameter
        """
        result = [
            
        ]
        pass

    def Transpose(self):
        """
        This function transposes the matrix.
        """
        """
        cols = [self.get_column(i) for i in range(self.cols)]
        return Matrix(cols)
        """
        transposed_data = [
            [self.arrays[j][i] for j in range(self.rows)]
            for i in range(self.cols)
        ]
        return Matrix(transposed_data)
    
    def is_SquareMatrix(self):
        """
        A square matrix is a matrix with the same number of rows and columns. 
        It occupies an important place, 
        especially in the definition of determinants
        """

        if self.rows == self.cols:
            return True 
        else: 
            return False
        
    def is_SymmetricalMatrix(self):
        """
            A matrix is symmetrical if it is equal to its transpose. A == Aᵗ
        """
        if self.is_SquareMatrix == False:
            return False
    
        transposed = self.Transpose()
        if self == transposed:
            return True
        
    """def Cofactor(self, rows , cols):
        first_product = (-1)**(rows+cols)

        cofactor_matris = 

        second_product = 
        


        pass"""
        
    def Determinant(self):
        """
        In mathematics, the determinant is a scalar-valued function of the entries of a square matrix.
        The determinant of a matrix A is commonly denoted det(A),
        det A, or |A|. Its value characterizes some properties of the matrix and the linear map represented, 
        on a given basis, by the matrix.
        """

        if self.is_SquareMatrix() == False:
            raise DefinitionError("The determinant is defined in square matrices")
        try:
            if self.FindDimension() == (2,2):
                value = self.FindElement(1,1)*self.FindElement(2,2) - self.FindElement(1,2)*self.FindElement(2,1)
                return value
            
            if self.FindDimension() == (3,3):
                """
                sarrus method
                """
                right = (
                self.FindElement(1,1)*self.FindElement(2,2)*self.FindElement(3,3)+
                self.FindElement(2,1)*self.FindElement(3,2)*self.FindElement(1,3)+
                self.FindElement(3,1)*self.FindElement(1,2)*self.FindElement(2,3)
                )

                left = (
                self.FindElement(3,1)*self.FindElement(2,2)*self.FindElement(1,3)+
                self.FindElement(2,3)*self.FindElement(3,2)*self.FindElement(1,1)+
                self.FindElement(3,3)*self.FindElement(1,2)*self.FindElement(2,1)
                )

                return right + left
        except TypeError:
            raise NonSymbolicValue("This determinant function can only calculate on numeric data. For symbolic data, please refer to the SymbolicMatrix function.")


    def FindElement(self, row, cols):
        """
        It allows to find the element in the desired column and row.
        """
        try:
            result = self.arrays[row-1][cols-1]
        except IndexError:
            raise DefinitionError("More value were entered than the number of rows or columns in the matrix.")

        return result
    
    def FindDimension(self):
        """
        Returns the dimensions of the input matrix

        Example: 
            >>> Matrix.FindDimension(1,2)
                search (a)1,2
                and finding in matrix

        """
        return (self.rows ,self.cols)


    def __matmul__(self,other):
        """
        classic row-by-column and column-by-row multiplication
        """
        if not self.cols == self.rows:
            raise DefinitionError("In order to multiply two matrices (A@B), the number of columns of matrix A must be equal to the number of rows of matrix B.")
    
        result = [[0 for _ in range(len(other.arrays[0]))] for _ in range(len(self.arrays))]

        for i in range(len(self.arrays)):         
            for j in range(len(other.arrays[0])):   
                for k in range(len(other.arrays)):  
                    result[i][j] += self.arrays[i][k] * other.arrays[k][j]

        return Matrix(result)

    def __mul__(self , other):
        """
        Hadamard product 
        """
        if self.rows != other.rows or self.cols != other.cols:
            raise DefinitionError("The number of rows and columns of two matrices must be equal")
        
        try:
            result = [
                [self.arrays[i][j] * other.arrays[i][j] for j in range(self.cols)]
                for i in range(self.rows)
            ]
            return Matrix(result)
        except IndexError:
            raise DefinitionError("wrong index defined")
        



    """  
    def __eq__(self, other):
        if self.rows != other.rows or self.cols != other.cols:
            raise ValueError("The number of rows and columns in the matrix must be the same")
        for i in 
    """


    def __eq__(self, other):
        if self.rows != other.rows or self.cols != other.cols:
            return False
        for i in range(self.rows):
            for j in range(self.cols):
                if self.arrays[i][j] != other.arrays[i][j]:
                    return False
        return True

    def __add__(self, other):
        """
        Necessary definitions and operations for matrix addition
        """
        if self.rows != other.rows or self.cols != other.cols:
            raise ValueError("The number of rows and columns in the matrix must be the same")

        try:
            result = [
                [self.arrays[i][j] + other.arrays[i][j] for j in range(self.cols)]
                for i in range(self.rows)
            ]
            return Matrix(result)
        except IndexError:
            raise DefinitionError("For the aggregation to be defined, the number of rows and columns must be the same.")

    def __sub__(self ,other):
        if self.rows != other.rows or self.cols != other.cols:
            raise ValueError("The number of rows and columns in the matrix must be the same")

        try:
            result = [
                [self.arrays[i][j] - other.arrays[i][j] for j in range(self.cols)]
                for i in range(self.rows)
            ]
            return Matrix(result)
        except IndexError:
            raise DefinitionError("For the subregation to be defined, the number of rows and columns must be the same.")

    def __repr__(self):
        return "\n".join(str(row) for row in self.arrays)

"""A = Matrix([
    [1, 2, 3],
    [2, 5, 5],
    [2, 5, "n"]
])

print(A.Determinant())"""



from .errors import *
import math

class Vector:
    def __init__(self, dimensions):
        if not isinstance(dimensions , list) :
            raise TypeError("You must list the dimensions for manipulation")
        if not 1<len(dimensions)<=3:
            raise ValueError("The vectors you define can be 2 or 3 dimensional")
        self.dimensions = dimensions

    def magnitude(self):
        """
        function to find the magnitude of the vector
        """
        if len(self.dimensions) == 2:
            x = self.dimensions[0] 
            y = self.dimensions[1]
            length = (x**2 + y**2)**0.5

        elif len(self.dimensions) == 3:
            x = self.dimensions[0] 
            y = self.dimensions[1]
            z = self.dimensions[2]
            length = (x**2 + y**2 + z**2)**0.5
        return length

    def information(self):
        """
        This function returns the information of the defined vector.
        """
        vektorinformation = [

        ]
        if len(self.dimensions) == 2:
            x = self.dimensions[0] 
            y = self.dimensions[1]
            vektorinformation.append({"dimension":2 })
            length = (x**2 + y**2)**0.5
            vektorinformation.append(length)

        elif len(self.dimensions) == 3:
            x = self.dimensions[0] 
            y = self.dimensions[1]
            z = self.dimensions[2]
            vektorinformation.append({"dimension":2})
            length = (x**2 + y**2 + z**2)**0.5
            vektorinformation.append({"length":length})    
        return vektorinformation
    
    def Inner_product(self, other):
        """
        This function finds the inner product of two defined vectors.
        """

        if len(self.dimensions) != len(other.dimensions):
            raise ValueError("Vectors must have the same dimension for dot product.")
        return sum(a*b for a, b in zip(self.dimensions, other.dimensions))
    
    def AngleTwoVector(self, other):
        """
        It is a function that finds the angle between two vectors according to the inner product equation.
        """
        if len(self.dimensions) != len(other.dimensions):
            raise ValueError("Vectors must have the same dimension for dot product.")

        inner_product = self.Inner_product(other)
        mag1 = self.magnitude()
        mag2 = other.magnitude()

        cos_theta = max(-1, min(1, inner_product / (mag1 * mag2)))

        return math.acos(cos_theta)
    
    def Visualization(self):
        try: 
            import matplotlib.pyplot as plt
        except ModuleNotFoundError:
            raise RequiredModule(
            "matplotlib module not installed. Install it with 'pip install matplotlib'."
            )

        v = self.dimensions

        if len(v) == 2:
            origin = Vector([0, 0])
            plt.figure(figsize=(6,6))
            plt.quiver(*origin, *v, angles='xy', scale_units='xy', scale=1, color='r', width=0.01)
            plt.xlim(-1, self.dimensions[0]+1)
            plt.ylim(-1,self.dimensions[1]+1)
            plt.grid(True)
            plt.axhline(0, color='black', linewidth=0.8)
            plt.axvline(0, color='black', linewidth=0.8)
            plt.text(v[0]/2, v[1]/2, f"{v}", fontsize=12, color='blue')

            plt.title("R^2 Visualization")
            plt.xlabel("x axis")
            plt.ylabel("y axis")
            plt.gca().set_aspect('equal', adjustable='box')
            plt.show()
        elif len(v) == 3:
            origin = Vector([0, 0, 0])
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')

            ax.quiver(origin[0], origin[1], origin[2],
                    v[0], v[1], v[2],
                    color='b', arrow_length_ratio=0.1)

            ax.set_xlim([-1, self.dimensions[0]+1])
            ax.set_ylim([-1, self.dimensions[1]+1])
            ax.set_zlim([0, self.dimensions[2]+1])

            ax.set_xlabel('X axis')
            ax.set_ylabel('Y axis')
            ax.set_zlabel('Z axis')
            ax.set_title('R^3 Visualization')

            plt.show()

    def UnitVector(self):
        """
        Returns the unit vector of the given vector.
        """
        Norm = self.magnitude()
        if Norm == 0:
            raise ValueError("Zero is undefined in division")
        return [x / Norm for x in self.dimensions]
    
    def CosTwoVector(self, other):
        """
        Finds the cosine of the angle between two vectors.
        cos(θ) = (v1 · v2) / (‖v1‖ * ‖v2‖)
        """

        InnerP = self.Inner_product(other)

        normv1 = self.magnitude()
        normv2 = other.magnitude()

        if normv1 == 0 or normv2 == 0:
            raise ValueError("Zero vector: cosine undefined.")

        return InnerP / (normv1 * normv2)
    
    def Projection(self, other):
        """
        Returns the projection of one vector onto another vector.
        proj_v2(v1) = (|v1| * cosθ) * unit(v2)
        """
        unitv2 = other.UnitVector()
        cosa = self.CosTwoVector(other)
        normv1 = self.magnitude()

        scalar = normv1 * cosa
        return [scalar * x for x in unitv2]

    def __repr__(self):
        return f"Vector({self.dimensions})"

    def __str__(self):
        if len(self.dimensions) == 2:
            dimensions = "R^2"
        elif len(self.dimensions) == 3:
            dimensions = "R^3"

        return f"<Vector {self.dimensions}, |v|={self.magnitude():.2f} , dimensions={dimensions}>"
        
    def __add__(self, other):
        """Vector addition"""
        if len(self.dimensions) != len(other.dimensions):
            raise ValueError("Vectors must have the same dimension for addition.")
        return Vector([a + b for a, b in zip(self.dimensions, other.dimensions)])

    def __sub__(self, other):
        """Vector subtraction"""
        if len(self.dimensions) != len(other.dimensions):
            raise ValueError("Vectors must have the same dimension for subtraction.")
        return Vector([a - b for a, b in zip(self.dimensions, other.dimensions)])

    def __mul__(self, scalar):
        """Scalar multiplication"""
        if not isinstance(scalar, (int, float)):
            raise TypeError("You can only multiply a vector by a scalar (number).")
        return Vector([a * scalar for a in self.dimensions])

    def __matmul__(self, other):
        """Dot product with @ operator"""
        return self.Inner_product(other)

    def __eq__(self, other):
        """Check if two vectors are equal"""
        return self.dimensions == other.dimensions
    
    def __getitem__(self, index):
        return self.dimensions[index]

    def __iter__(self):
        return iter(self.dimensions)
    
class Line:
    def __init__(self, point, direction=None, slope=None, perpendicular_vector=None):
        if not (isinstance(point, list) and isinstance(direction, list)):
            raise WrongDataTypeError("Must be in the point and direction list type")
        if len(point) != len(direction):
            raise SizeLimitExceededError("Point and direction list must be the same size")
        if len(point) not in (2, 3):
            raise SizeLimitExceededError("Only dimensions 2 and 3 are supported")

        self.point = point
        #self.direction = direction

        if slope is None:
            if len(point) == 2:
                dx = direction[0]
                dy = direction[1]
                self.slope = None if dx == 0 else (dy / dx)
            else:
                self.slope = None
        else:
            if not isinstance(slope, (int, float)):
                raise WrongDataTypeError("Slope must be float")
            self.slope = float(slope)

        if perpendicular_vector is None:
            if len(direction) == 2:
                a, b = direction
                self.perpendicular_vector = [-b, a]
            else:
                self.perpendicular_vector = None
        else:
            if not isinstance(perpendicular_vector, list):
                raise WrongDataTypeError("perpendicular_vector must be list")
            self.perpendicular_vector = perpendicular_vector


    def point_at(self, t):
        if not isinstance(t, (int, float)):
            raise WrongDataTypeError("t must be float")
        return [p + t * d for p, d in zip(self.point, self.direction)]
    
    def __repr__(self):
        return self.__str__()
    
    def __str__(self):
        return f"r(t) {self.point} t*{self.direction}"
        


from .errors import NonCompliancaRecognition
from itertools import count 


def FiniteCrateSet(i,n,k) -> list:

    """
    This function allows you to create a cluster

    Parametres:
        i : start value 
        n : stop value 
        k : multiples 

    Examples: 
        i = 1 , n = 14 , k = 2
        {2,4,6,8,10,12,14}
    
    """
    for j in range(i,n+1):
        sets = list()
        if j % k == 0 :
            sets.append(j)
    return sets

#"""def Natural():
#    """
    #ℕ :
    #This function represents the set of natural numbers.
    #According to Peano's axioms and ours, 0 is not included.
    
    #This function continues to translate until it cannot be stopped under the required conditions.

#    {1,2,3,...}
    """

    Naturals = Mset(1,"inf",LowOpenRange=False ,HighOpenRange=False , Excluded=None)
    
    return Naturals
"""
        
def Bool_Natural(n: int) -> bool:
    """
    ℕ :
       This function checks whether the parameter is a natural number.
    
    Return:
        True , False
    """

    if n >= 1 and n.is_integer() == True:
        return True
    else:
        False

def Bool_Integer(n:int): 
    """
    ℤ:
        This function checks whether the parameter is an integer.
    Return:
        True , False
    """
    
    return n.is_integer()

def Integer(n: int) :
    """
    ℤ:
    This function contains the set of integers defined by mathematics.
    The set of integers is not a countably infinite set.

    {..., -2, -1, 0, 1, 2, ...}
    """
    n = 0
    yield n
    step = 1
    while True:
        yield step
        yield -step
        step += 1

def Bool_RealNumber(n: int):
    """
    Real numbers most popular number sets in math.
    Real numbers are arguably the most fundamental and commonly used set of numbers in mathematics.
    Simply put,
    a real number is any value that can represent a distance along a continuous line (the number line).
    
    Parameters: 
        n : control number 
    
    Examples: 
    >>>    for n = 1/2 
        Bool_RealNumber(1/2) = True
    >>> for n = 3 - 4i 
        Bool_RealNumber(n) = True

    """
    if isinstance(n, (int, float)):
        return True
    elif isinstance(n, complex):
        return n.imag == 0  # Sanal kısmı yoksa reel kabul edilir
    else:
        return False


#This file allows you to define data types that do not exist in Python.


class Mset:
    def __init__(self, LowLimit=None , HighLimit=None, LowOpenRange=False ,HighOpenRange= False, Excluded=None):

        """
        This class defines the set mathematically.
        """
        
        self.LowLimit = LowLimit
        self.HighLimit = HighLimit
        self.LowOpenRange = LowOpenRange
        self.HighOpenRange = HighOpenRange
        self.Excluded = set(Excluded or [])

    def __contains__(self, x):
        if self.LowLimit is not None:
            if self.LowOpenRange and x <= self.LowLimit:
                return False
            if not self.LowOpenRange and x < self.LowLimit:
                return False
        
        if self.HighLimit is not None:
            if self.HighOpenRange and x >= self.HighLimit:
                return False
            if not self.LowOpenRange and x > self.HighLimit:
                return False

        if x in self.Excluded:
            return False

        return True
    
    @staticmethod
    def from_string(s):
        s = s.strip().replace(" ", "")  
        
        range_part, *excluded_part = s.split("-")
        
        acik_alt = range_part.startswith("(")
        acik_ust = range_part.endswith(")")
        
        range_numbers = range_part[1:-1].split(",")
        alt = float(range_numbers[0])
        ust = float(range_numbers[1])
        
        if excluded_part:
            excluded_str = excluded_part[0].strip("{}")
            excluded = set(map(float, excluded_str.split(","))) if excluded_str else set()
        else:
            excluded = set()

        return Mset(alt, ust, acik_alt, acik_ust, excluded)

    def __repr__(self):

        low = "(" if self.LowOpenRange else "["
        high = ")" if self.HighOpenRange else "]"
        if self.Excluded == None:
            return f"{low}{self.LowLimit}, {self.HighLimit}{high}"
        else:
            return f"{low}{self.LowLimit}, {self.HighLimit}{high} - {{{', '.join(map(str, sorted(self.Excluded)))}}}"


from .errors import RequiredModule


try:
    from sympy import symbols, Function, simplify, Matrix as SymMatrix
except ModuleNotFoundError:
    raise RequiredModule("The unimath library works integrated with sympy to perform symbolic operations, so we use sympy in functions that require symbolic operations.")



class SymbolicVariable:
    """
    Sembolik değişkenleri temsil eder. Örn: n, k, x
    """
    def __init__(self, name: str):
        self.symbol = symbols(name)

    def __repr__(self):
        return f"SymbolicVariable({self.symbol})"


class SymbolicExpression:
    """
    Her türlü sembolik ifadeyi taşır.
    Toplama, çıkarma, çarpma gibi işlemler overload edilmiştir.
    """
    def __init__(self, expr):
        self.expr = expr

    def simplify(self):
        return SymbolicExpression(simplify(self.expr))

    def __add__(self, other):
        return SymbolicExpression(self.expr + other.expr)

    def __sub__(self, other):
        return SymbolicExpression(self.expr - other.expr)

    def __mul__(self, other):
        return SymbolicExpression(self.expr * other.expr)

    def __truediv__(self, other):
        return SymbolicExpression(self.expr / other.expr)

    def __repr__(self):
        return f"SymbolicExpression({self.expr})"


class Sequence:
    """
    Sembolik diziler (a_n) oluşturmak için.
    rule: lambda n: ifade
    """
    def __init__(self, rule):
        self.rule = rule
        self.n = symbols("n", integer=True)

    def term(self, k: int):
        return self.rule(k)

    def symbolic_term(self):
        return self.rule(self.n)
    
    def monotony(self):
        """
        the basic rule of monotony:
        >>> (a)n+1 - (a)n > 0 -> increasing
        >>> (a)n+1 - (a)n < 0 -> decreasing
        >>> (a)n+1 - (a)n = 0 -> constant
        """
        
        if simplify(self.rule(self.n+1)/self.rule(self.n)) > 1:
            return "increasing"
        if simplify(self.rule(self.n+1)/self.rule(self.n)) < 1:
            return "decreasing"
        if simplify(self.rule(self.n+1)/self.rule(self.n)) == 1:
            return "constant"


class SymbolicMatrix:
    """
    Normal matris sınıfınıza sembolik destek için bir wrapper.
    """
    def __init__(self, data):
        self.data = data
        self.matrix = SymMatrix(data)

    def determinant(self):
        return simplify(self.matrix.det())

    def transpose(self):
        return SymbolicMatrix(self.matrix.T.tolist())

    def __repr__(self):
        return f"SymbolicMatrix({self.matrix})"



# for sup and inf function
 
from unimath.mset import Mset
from parse import ParseMset

def sup(set: Mset) -> Mset:

    parsed = set.ParseMset(set)

    NLowLimit = parsed.get("Lowlimit", None)
    NLowOpenRange = parsed["LowOpenRange"]


    if NLowOpenRange == False and NLowOpenRange == True:
        return NLowLimit
    

def inf(set: Mset) -> Mset:

    parsed = set.ParseMset(set)

    NHighLimit = parsed.get("HighLimit", None)
    NHighOpenRange = parsed["HighOpenRange"]


    if NHighOpenRange == False and NHighOpenRange == True :
        return NHighLimit




# Poyraz Soylu
# unimath
# analysis functions

from functools import reduce
import operator


def sigmanotation(i: int , n:int , func) -> int:
    """
    Σ (Sigma Notation) - Summation

    Calculates the sum of a sequence of terms from start to end.
    This function generalizes the idea of summation using a custom function.

    Parameters:
        func (callable): A function f(i) that defines the terms of the sequence.
        i (int): The starting index (inclusive).
        n (int): The ending index (inclusive).

    Returns:
        int or float: The sum of f(i) for i in [i,n].

    Example:
        >>> sigmanotation(lambda i: i, 1, 5)
        15  # (1 + 2 + 3 + 4 + 5)

        >>> sigmanotation(lambda k: k**2, 1, 4)
        30  # (1^2 + 2^2 + 3^2 + 4^2)
    """
    for i in range(i,n + 1):
        values = list()
        values.append(func(i))

        return sum(values)


def productnatation(i: int, n:int , func) -> int:
    """
    Π (Product Notation) - Multiplication of terms

    Calculates the product of a sequence of terms from start to end.
    This function generalizes the idea of multiplication using a custom function.

    Parameters:
        func (callable): A function f(i) that defines the terms of the sequence.
        i (int): The starting index (inclusive).
        n (int): The ending index (inclusive).

    Returns:
        int or float: The product of f(i) for i in [i,n].

    Example:
        >>> productnatation(lambda i: i, 1, 4)
        24  # (1 * 2 * 3 * 4)

        >>> productnatation(lambda k: k/(k+1), 1, 3)
        0.25  # (1/2 * 2/3 * 3/4)
    """
    return reduce(operator.mul, (func(i) for i in range(i, n + 1)), 1)

def factorial(n: int) -> int:
    """
    Calculates the factorial of a non-negative integer n.

    Factorial (n!) is the product of all positive integers from 1 to n.
    It is widely used in combinatorics, probability, and algebra.

    Formula:
        n! = n * (n-1) * (n-2) * ... * 1
        0! = 1 (by definition)

    Parameters:
        n (int): A non-negative integer

    Returns:
        int: The factorial of n

    Raises:
        ValueError: If n is negative
    """
    try:
        result = 1
        for i in range(1, n + 1):
            result *= i
        return result
    except ValueError:
        print("[ERROR] Factorial is not defined for negative numbers.")

"""def Transformation(rule , domain ):
    parsed = ParseMset(str(domain))

    NLowLimit = parsed.get("Lowlimit", None)
    NHighLimit = parsed.get("HighLimit", None)
    NLowOpenRange = parsed["LowOpenRange"]
    NHighOpenRange = parsed["HighOpenRange"]
    NExcluded = parsed["Excluded"]

    return Mset(
        rule(float(NLowLimit)),
        rule(float(NHighLimit)),
        NLowOpenRange,
        NHighOpenRange,
        NExcluded
    )"""



from .calculuscalculus import factorial
from .errors import NonCompliancaRecognition 

def permutation(repetetion:str, n:int, r:int) ->int :    
    """
    In mathematics, a permutation of a set can mean one of two different things:

    an arrangement of its members in a sequence or linear order, or
    the act or process of changing the linear order of an ordered set.

    Formula: 
        if repetetion is 1 = n!/(n-r)!
        receptions is 0 = n^r

    Parameters:
        repetetion(int)= whether the permutation is repeated or not
        n(int)
        r(int)

    Return:
        return a permutation of given values
    """
        #terms
    if n>=r:
        pass
    else:
        print("[ERROR] Does not fit this description")
        exit()

    if repetetion == "0":
        return factorial(n)/factorial(n - r)
    elif repetetion == "1":
        return n**r
    
    
def combination(n: int, r: int) -> int: 
    """
    In mathematics,
    a combination is a selection of items from a set that has distinct members,
    such that the order of selection does not matter

    Formula: 
        c(n,r) = n!/r!(n-r)!

    Parametres:
        n(int)
        r(int)
    
    Return:
        return a combination of given values

    """
    #terms
    if n>=r:
        pass
    elif n == r:
        return 1
    else:
        NonCompliancaRecognition()

    return factorial(n)/factorial(r)*factorial(n - r)

def posibilty(n:int , r: int) -> int:
    """
    Probability is a branch of mathematics and statistics concerning events and numerical
    descriptions of how likely they are to occur.
    The probability of an event is a number between 0 and 1;
    the larger the probability, the more likely an event is to occur.

    Formula:
        
    """
    if r == 0 and n>r :
        NonCompliancaRecognition()
    
    return n/r




#the file used to perform the parse operation


import sys, os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

from unimath.mset import Mset

def ParseMset(s):
    """
    ParseMset:
    This function parses the defined set.

    Parametres:
    s : defined set

    """
    s = s.strip().replace(" ", "") 
    
    range_part, *excluded_part = s.split("-")
    
    LowOpenRange = range_part.startswith("(")
    HighOpenRange = range_part.endswith(")")
    
    range_numbers = range_part[1:-1].split(",")
    LowLimit = float(range_numbers[0])
    HighLimit = float(range_numbers[1])
    
    if excluded_part:
        excluded_str = excluded_part[0].strip("{}")
        excluded = set(map(float, excluded_str.split(",")))
    else:
        excluded = set()
    
    return {
        "LowLimit": LowLimit,
        "HighLimit": HighLimit,
        "LowOpenRange": LowOpenRange,
        "HighOpenRange": HighOpenRange,
        "Excluded": excluded
    }



from setuptools import setup, find_packages

setup(
    name="unimath",
    version="1.4.3",
    packages=find_packages(),
    install_requires=[
        "sympy","matplotlib"
    ],
    author="Poyraz Soylu",
    author_email="psoylu3467@gmail.com",
    description="Mathematical Environment Library in Python",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Jolankaa/unimath",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)